#include	<stdlib.h>
#include	<string.h>

#include	<yz/window.h>
#include	<yz/draw.h>
#define		YZ_UITK_IMPLEMENTOR
#include	"uitk.h"

#include	"button.h"
#include	"private.h"

void					yzDrawButton(button_t	*b);
int						point_in_button(int	x, int	y, button_t	*b);
int						button_filter(event_t	*ev);

static int					buttonType = 0;
#define	TYPE_NAME	"button"
static ui_element_type_info	typeInfo = { TYPE_NAME, 0, 0, button_filter };

#define	CHECK_REG()	if(register_button_type())return 0;

register_button_type()
{
	buttonType = yzUIRegisterTypeInfo(&typeInfo);
	if(buttonType==0)return -1;
	return 0;
}

button_t	*yzAddButton(char	*name, int	x, int	y, int	dx, int	dy, int	id)
{
	button_t	*b;
	
	CHECK_REG();
	b = (button_t*)malloc(sizeof(button_t));
	b->type = buttonType; b->font = 0;
	b->x = x; b->y = y; b->dx = dx; b->dy = dy;
	strcpy(b->name, name); b->id = id;
	b->click_handler = 0; b->state = 0;
	b->key_equiv = -1;
	yzUIAddToList(b);
	yzDrawButton(b);
	return b;
}

button_t	*yzAddButtonWithHandler(char	*name, int	x, int	y, int	dx, int	dy,
						int	id, void	(*click_handler)(int))
{
	button_t	*b;
	
	CHECK_REG();
	b = (button_t*)malloc(sizeof(button_t));
	b->type = buttonType; b->font = 0;
	b->x = x; b->y = y; b->dx = dx; b->dy = dy;
	strcpy(b->name, name); b->id = id;
	b->click_handler = click_handler; b->state = 0;
	b->key_equiv = -1;
	yzUIAddToList(b);
	yzDrawButton(b);
	return b;
}

void		yzRemoveButton(button_t	*b)
{
	if(!b)return;
	yzUIRemoveFromList(b);
	if(b->font)free(b->font);
	free((char*)b);
}

void	yzSetButtonFont(button_t	*b, char	*font)
{
	if(!b)return;
	if(!font)b->font = 0;
	else
	{
		if(b->font)free(b->font);
		b->font = malloc(strlen(font)+1);
		strcpy(b->font, font);
	}
	yzDrawButton(b);
}

void	yzSetButtonKeyEquiv(button_t	*b, int	key)
{
	if(!b)return;
	b->key_equiv = key;
}

int		button_filter(event_t	*ev)
{
	ui_element_list	*lst, *i;
	button_t		*b;
	
	lst = yzUIGetList(TYPE_NAME);
	switch(ev->type)
	{
		case	YZ_KEY_DOWN:
			for(i=lst;i;i=i->next)
			{
				b = i->e;
				if(b->key_equiv != -1 && ev->key == b->key_equiv)
				{
					ev->type = YZ_BUTTON_HIT;
					ev->button = b->id;
					return 0;
				}
			}
			return 0;
		case	YZ_MOUSE_DOWN:
			for(i=lst;i;i=i->next)
			{
				b = i->e;
				if(point_in_button(ev->x, ev->y, b))
				{
					b->state = 1;
					yzDrawButton(b);
					return 1;
				}
			}
			return 0;
		case	YZ_MOUSE_DRAG:
			for(i=lst;i;i=i->next)
			{
				b = i->e;
				if(point_in_button(ev->x, ev->y, b))
				{
					if(b->state == -1)
					{
						b->state = 1;
						yzDrawButton(b);
					}
					if(b->state != 0)return 1;
				}
				else if(b->state == 1)
				{
					b->state = 0;
					yzDrawButton(b);
					b->state = -1;
				}
			}
			return 0;
		case	YZ_MOUSE_UP:
			for(i=lst;i;i=i->next)
			{
				b = i->e;
				if(b->state == 1) // mouse was down over this button
				{
					b->state = 0;
					yzDrawButton(b);
					// take notification action
					if(b->click_handler)
					{
						b->click_handler(b->id);
						return 1;
					}
					else
					{
						// return our customized button event
						ev->type = YZ_BUTTON_HIT;
						ev->button = b->id;
						return 0;
					}
				}
				else if(b->state == -1)	// mouse was down, not over the button (==canceled)
				{
					b->state = 0;
					return 1;
				}
			}
			return 0;
		default:	return 0;	// don't mess with most events.
	}
}
