#include	<stdio.h>
#include	<string.h>
#include	<stdlib.h>
#include	<sys/time.h>

#include	<yz/window.h>
#include	<yz/draw.h>
#include	"menu.h"

menu_data_t		*yzNewMenuData(char	*name)
{
	menu_data_t	*menu;
	
	menu = (menu_data_t*)malloc(sizeof(menu_data_t));
	if(!menu)return 0;
	
	menu->name = malloc(strlen(name)+1);
	if(!menu->name)
	{
		free((char*)menu);
		return 0;
	}
	strcpy(menu->name, name);
	menu->nitems = 0;
	menu->items = 0;
	
	return menu;
}

void	yzDisposeMenuData(menu_data_t	*m)
{
	int	i;
	
	if(!m)return;
	for(i=0;i<m->nitems;i++)free(m->items[i].name);
	if(m->items)free((void*)m->items);
	free((void*)m);
}

int	yzAddMenuDataItem(menu_data_t	*menu, char	*item, int	id)
{
	menu_item_t	*ni;
	
	ni = realloc(menu->items, sizeof(menu_item_t) * (menu->nitems+1));
	if(!ni)return -1;
	else menu->items = ni;
	menu->items[menu->nitems].name = malloc(strlen(item)+1);
	if(!menu->items[menu->nitems].name)return -1;
	strcpy(menu->items[menu->nitems].name, item);
	menu->items[menu->nitems].id = id;
	menu->nitems++;
	return 0;
}

window_t	*open_borderless_window(int	x, int	y, int	width, int	height)
{
	int			cx, cy;
	
	yzGetWindowCorner(&cx, &cy);
	//printf("window is at %d, %d\n", cx, cy);
	x += cx;
	y += cy;
	// 32 means borderless
	return (window_t*)yzNewWindowFlagsAt("Menu Window",
							x, y, width, height, 32);
}

typedef struct milliclock { struct timeval	tv; char pad[32 - sizeof(struct timeval)]; } milliclock;

void	start_milliclock(milliclock	*cl)
{
	struct timezone	tz;
	
	gettimeofday(&cl->tv, &tz);
}

int	millitime(milliclock	*cl)
{
	struct timeval	tv;
	struct timezone	tz;
	int		millis;
	
	gettimeofday(&tv, &tz);
	millis = (tv.tv_usec - cl->tv.tv_usec)/1000;
	millis += 1000 * (tv.tv_sec - cl->tv.tv_sec);
	return millis;
}

int	yzShowMenu(menu_data_t	*m, int	*selid, int	mx, int	my)
{
	window_t	*menu, *old;
	event_t		ev;
	int			width, height, i, sel = -4, r;
	milliclock	mc;
	
	old = yzGetSelectedWindow();
	start_milliclock(&mc);
	width = 120 + 6;
	height = m->nitems * 16 + 6;
	menu = open_borderless_window(mx-10, my-10, width, height);
	yzSelectWindow(menu);
	yzPenColor(40000, 33000, 50000);
	yzClear();
	yzPenColor(25000, 25000, 32500);
	yzRect(1, 1, width-2, height-2);
	yzPenColor(12000, 12000, 15000);
	yzRect(2, 2, width-3, height-3);
	yzPenColor(0, 0, 0);
	yzRect(3, 3, width-4, height-4);
	for(i=0;i<m->nitems;i++)
		draw_menu_item(3, (i)*16 + 3, width-6, 16, &m->items[i], 0);

	while((r=yzQueryPointer(&mx, &my)) < 3 || millitime(&mc)<200)
	{
		ev = yzGetEvent(50);
		if(ev.type == YZ_WINDOW_DEACTIVATE)break;
		if(ev.type == YZ_MOUSE_UP)break;
		my -= 3;
		mx -= 3;
		if(my<0)my=-100;
		my /= 16;
		//printf("mouse item = %d, sel item = %d, r %d\n", my, sel, r);
		if(my != sel || mx<-2 || mx>(width-3))
		{
			if(sel>=0)draw_menu_item(3, (sel)*16 + 3, width-6, 16, &m->items[sel], 0);
			if(my>=0 && my<m->nitems && mx>=0 && mx<(width-6))
			{
				sel = my;
				draw_menu_item(3, (sel)*16 + 3, width-6, 16, &m->items[sel], 1);
			}
			else sel = -1;
		}
	}
	yzDeleteWindow(menu);
	yzSelectWindow(old);
	if(selid)
	{
		if(sel>=0)*selid = m->items[sel].id;
		else *selid = 0;
	}
	return sel;
}

draw_menu_item(int	x, int	y, int	dx, int	dy, menu_item_t	*it, int	s)
{
	if(!s)yzPenColor(45000, 45000, 60000);
	else yzPenColor( 30000, 30000, 60000);
	yzRect(x, y, x+dx-1, y+dy-1);
	yzPenColor( 15000, 15000, 10000);
	yzText(x+2, y+14, it->name);
}
