#include	<stdio.h>
#include	<string.h>
#include	<stdlib.h>
#include	<sys/time.h>
#include	<stdarg.h>

#include	<yz/window.h>
#include	<yz/draw.h>

#define		YZ_UITK_IMPLEMENTOR
#include	"uitk.h"
#include	"menu.h"

int		popup_menu_filter(event_t	*ev);
void	draw_popup_menu(popup_menu_t	*m, int	hilite);
int	yzShowMenu(menu_data_t	*m, int	*selid, int	mx, int	my);

#define	TYPE_NAME	"popup-menu"

static ui_element_type_info	typeInfo = {TYPE_NAME, 0, 0, popup_menu_filter};
static int						popupMenuType = 0;

#define	CHECK_REG()	if(register_popup_menu_type())return 0;

register_popup_menu_type()
{
	popupMenuType = yzUIRegisterTypeInfo(&typeInfo);
	if(popupMenuType==0)return -1;
	return 0;
}

popup_menu_t		*yzAddPopupMenuFromMenuData(menu_data_t	*m, int	x, int	y,
									int	dx, int	dy, int	id)
{
	popup_menu_t	*pm;
	
	if(!m)return 0;
	CHECK_REG();
	pm = (popup_menu_t*)malloc(sizeof(popup_menu_t));
	if(!pm)return 0;
	
	pm->type = popupMenuType;
	pm->id = id;
	pm->own_menu = 0;
	pm->menu = m;
	pm->x = x; pm->y = y;
	pm->dx = dx; pm->dy = dy;
	pm->lastSel = 0;
	
	yzUIAddToList(pm);
	draw_popup_menu(pm, 0);
	
	return pm;
}

void	yzRemovePopupMenu(popup_menu_t	*m)
{
	int	i;
	
	if(!m)return;
	yzUIRemoveFromList(m);
	if(m->own_menu)yzDisposeMenuData(m->menu);
	free((void*)m);
}

popup_menu_t	*yzAddPopupMenu(char	*name, int	x, int	y,
									int	dx, int	dy, int	id, ...)
{
	va_list	va;
	int		iid = 1;
	char	*iname;
	popup_menu_t	*pm;
	menu_data_t		*md;
	
	md = yzNewMenuData(name);
	va_start(va, id);
	while(iname = va_arg(va, char*))
		yzAddMenuDataItem(md, iname, iid++);
	va_end(va);
	pm = yzAddPopupMenuFromMenuData(md, x, y, dx, dy, id);
	pm->own_menu = 1;
	return pm;
}

int				yzPopupUIType(void) { return popupMenuType; }

char			*yzPopupGetItemName(popup_menu_t	*pm, int	i)
{
	if(i<0 || i>=pm->menu->nitems)return 0;
	return pm->menu->items[i].name;
}

void	draw_popup_menu(popup_menu_t	*m, int	hilite)
{
	int	xc, tw, xl, yc;
	
	yzPenColor(0, 0, 0);
	yzSelectFont("system");
	yzOutlineRect(m->x, m->y, m->x+m->dx-1, m->y+m->dy-1);
	if(!hilite)yzPenColor(50000, 50000, 50000);
	else yzPenColor(40000, 40000, 50000);
	yzRect(m->x+1, m->y+1, m->x+m->dx-2, m->y+m->dy-2);
	yzPenColor(0, 0, 0);
	xc = m->x + m->dx/2;
	xl = m->x + m->dx*0.85;
	yc = m->y + m->dy/2;
	tw = yzStringSize(m->menu->name, strlen(m->menu->name));
	if(xl>(xc+tw/2))
	{
		yzText(xc-tw/2 - 5, m->y+m->dy/2+5, m->menu->name);
		yzPolygon(3, xl, yc-5, xl+8, yc, xl, yc+5);
	}
	else yzText(xc-tw/2, m->y+m->dy/2+5, m->menu->name);
}

int	popup_menu_filter(event_t	*ev)
{
	ui_element_list	*lst, *i;
	popup_menu_t	*p;
	
	lst = yzUIGetList(TYPE_NAME);
	if(!lst)return 0;
	switch(ev->type)
	{
		case	YZ_MOUSE_DOWN:
			for(i=lst;i;i=i->next)
			{
				p = i->e;
				if(ev->x>=p->x && ev->y>=p->y &&
					ev->x<=p->x+p->dx && ev->y<=p->y+p->dy)
				{
					draw_popup_menu(p, 1);
					p->lastSel = yzShowMenu(p->menu, &p->lastSelID,
											p->x+15, p->y+15);
					draw_popup_menu(p, 0);
					ev->type = YZ_UI_EVENT;
					ev->ui_element = p;
					if(p->lastSel == -1)return 1;
					else return 0;
				}
			}
			break;
	}
	return 0;
}
