#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<time.h>
#include	<sys/time.h>

#include	<yz/window.h>
#include	<yz/draw.h>

#define	YZ_UITK_IMPLEMENTOR
#include	"uitk.h"
#include	"editline.h"

struct	editline_t
{
	int	type;	// always editlineType for an editline_t
	int	do_hide;	// if set, all chars show as *, i.e. password field
	int	x, y;
	int	dx, dy;
	char	text[512];
	int	selins, selto;	//selection start and end - if the same, = insertation point
	void	(*skip_handler)(struct editline_t*, int	key);
	editline_keyskips	skips;
};

#define	max(a, b)	((a>b)?(a):(b))
#define	min(a, b)	((a>b)?(b):(a))

int	editline_filter(event_t	*ev);
void	handle_editline_key(editline_t	*e, int	key);
void	handle_editline_click(editline_t	*e, int	x, int	y);
void	handle_editline_cursor_key(editline_t	*e, int	key);
void	yzDrawEditline(editline_t	*e);
void	erase_editline_selection(editline_t	*e);

static int	next_blink = 0, caret_vis = 0;

#define	WINDOW_HILITED	"window/hilited"	// XXX this hack needs to go... yuck.

static char	passwdtxt[257] = "********************************" \
						 "********************************" \
						 "********************************" \
						 "********************************" \
						 "********************************" \
						 "********************************" \
						 "********************************" \
						 "********************************";

static int					editlineType = 0;
#define	TYPE_NAME	"editline"
static ui_element_type_info	typeInfo = { TYPE_NAME, 0, 1, editline_filter };

#define	CHECK_REG()	if(register_editline_type())return 0;

register_editline_type()
{
	editlineType = yzUIRegisterTypeInfo(&typeInfo);
	if(editlineType==0)return -1;
	return 0;
}

int		centiseconds()
{
	static struct timeval	old_tv = {0,0};
	struct timeval	new_tv, diff_tv;
	struct timezone	tz;
	
	gettimeofday(&new_tv, &tz);
	if(old_tv.tv_usec == old_tv.tv_sec && old_tv.tv_sec==0)old_tv = new_tv;
	diff_tv.tv_sec = new_tv.tv_sec - old_tv.tv_sec;
	diff_tv.tv_usec = new_tv.tv_usec - old_tv.tv_usec;
	return (diff_tv.tv_sec*100 + diff_tv.tv_usec/10000);
}

void	yzFlashCaret()
{
	editline_t	*focus;
	
	if(!yzGetWindowData(WINDOW_HILITED))return;
	focus = (editline_t*)yzUIGetFocus();
	if(!focus || focus->type!=editlineType)return;
	
	if(focus->selins != focus->selto)return;	// these don't flash
	if(centiseconds() > next_blink)
	{
		next_blink = centiseconds() + 40;
		caret_vis = !caret_vis;
		if(!caret_vis)erase_editline_selection(focus);
		else draw_editline_selection(focus);
	}
}

void	yzEditlineHandleKey(editline_t	*e, int	key)
{
	if(key=='	' || key=='\n' || key==3)
	{
		return;	// we don't deal with these here - if this function was called,
			// it's safe to assume the client deals with tab and enter
	}
	else
	{
		if(e)
			handle_editline_key(e, key);
	}
}

editline_t	*yzAddEditline(char	*txt, int	x, int	y, int	dx, int	dy,
							int	id, editline_keyskips	skips)
{
	editline_t	*e;
	
	CHECK_REG();
	e = (editline_t*)malloc(sizeof(editline_t));
	e->type = editlineType;
	e->do_hide = 0;
	e->x = x; e->y = y;
	e->dx = dx; e->dy = dy;
	strcpy(e->text, txt);
	e->selins = strlen(txt); e->selto = strlen(txt);
	e->skip_handler = 0;
	e->skips = skips;
	yzUIAddToList(e);
	yzDrawEditline(e);
	return e;
}

editline_t	*yzAddEditlineWithHandlers(char	*txt, int	x, int	y, int	dx, int	dy,
										int	id, editline_keyskips	skips,
					void	(*skip_handler)(struct editline_t	*e, int	key))
{
	editline_t	*e;
	
	CHECK_REG();
	e = (editline_t*)malloc(sizeof(editline_t));
	e->type = editlineType;
	e->do_hide = 0;
	e->x = x; e->y = y;
	e->dx = dx; e->dy = dy;
	strcpy(e->text, txt);
	e->selins = strlen(txt); e->selto = strlen(txt);
	e->skip_handler = skip_handler;
	e->skips = skips;
	yzUIAddToList(e);
	yzDrawEditline(e);
	return e;
}

void	yzRemoveEditline(editline_t	*e)
{
	if(!e)return;
	yzUIRemoveFromList(e);
	if(e->do_hide)memset(e->text, 0, 500);	// blot out password fields
	free((char*)e);
}

char	*yzGetEditlineText(editline_t	*e)
{
	if(!e)return 0;
	return e->text;
}

void	yzSelectEditline(editline_t	*e, int	begin, int	end)
{
	editline_t	*focus;
	
	if(!e)return;
	// see if we need to dehilite the old focus
	focus = (editline_t*)yzUIGetFocus();
	if(focus && focus->type==editlineType)
		erase_editline_selection(focus);
	focus = e;
	caret_vis = 1;
	next_blink = centiseconds() + 35;
	focus->selins = begin;
	focus->selto = (end==-1)?(strlen(e->text)):(end);
	yzUIFocusItem((generic_ui_element_t*)focus);
	draw_editline_selection(focus);
}

void	yzDeselectEditline(editline_t	*e)
{
	if(!e)return;
	erase_editline_selection(e);
	if((editline_t*)yzUIGetFocus() == e)
		yzUIDefocusItem((generic_ui_element_t*)e);
}

editline_t	*yzGetSelectedEditline()
{
	editline_t	*e;
	
	e = (editline_t*)yzUIGetFocus();
	if(e && e->type==editlineType)return e;
	return 0;
}

int	editline_filter(event_t	*ev)
{
	ui_element_list	*l, *l0;
	editline_t	*e, *s;
	
	e = (editline_t*)yzUIGetFocus();
	if(e && e->type != editlineType)e = 0;
	l = yzUIGetList(TYPE_NAME);
	
	// handle any key events that pertain to the selected editline
	// or mouse events that may select another editline
	switch(ev->type)
	{
		case	YZ_KEY_DOWN:
			if(yzQueryKey(YZ_KEYCODE_CONTROL))return 0;	// don't mess with control combinations
			if(!e)return 0;	// no selected editline
			if(ev->key == '\t')	// tab
			{
				if(e->skips & YZ_EDITLINE_TAB)
				{
					if(e->skip_handler)
					{
						e->skip_handler(e, ev->key);
						return 1;
					}
					else return 0;
				}
				else
				{
					// find the list element
					l0 = l;
					while(l && l->e!=e)l=l->next;
					// first choice: there's a 'next' element in our list
					if(l && l->next)yzSelectEditline(l->next->e, 0, -1);
					// second choice: no 'next', set focus to the start of the list
					else if(l)yzSelectEditline(l0->e, 0, -1);
					// third choice: no list (!?); just refocus the current selection
					else yzSelectEditline(e, 0, -1);
					return 1;
				}
			}
			if(ev->key == '\n')	// return
			{
				if(e->skips & YZ_EDITLINE_RETURN)
				{
					if(e->skip_handler)
					{
						e->skip_handler(e, ev->key);
						return 1;
					}
					else return 0;
				}
			}
			if(ev->key == 3)	// enter
			{
				if(e->skips & YZ_EDITLINE_ENTER)
				{
					if(e->skip_handler)
					{
						e->skip_handler(e, ev->key);
						return 1;
					}
					else return 0;
				}
			}
			if(ev->key == 27)	// escape
			{
				if(e->skips & YZ_EDITLINE_ESCAPE)
				{
					if(e->skip_handler)
					{
						e->skip_handler(e, ev->key);
						return 1;
					}
					else return 0;
				}
			}
			// ev->key is a key that goes right into the normal flow of things
			handle_editline_key(e, ev->key);
			return 1;
		case	YZ_MOUSE_DOWN:
			while(l)
			{
				s = l->e;
				if(point_in_editline(ev->x, ev->y, s))break;
				l = l->next;
			}
			if(l)	//the click was in an editline
			{
				if(s != e) // it was not in the selected editline
					yzSelectEditline(s, s->selins, s->selto);
				handle_editline_click(s, ev->x, ev->y);
				return 1;
			}
			return 0;
		case	YZ_WINDOW_ACTIVATE:
			yzPutWindowData(WINDOW_HILITED, (void*)1);
			if(e)draw_editline_selection(e);
			return 0;
		case	YZ_WINDOW_DEACTIVATE:
			yzPutWindowData(WINDOW_HILITED, (void*)0);
			if(e)grey_editline_selection(e);
			return 0;
	}
	return 0;
}

void	handle_editline_key(editline_t	*e, int	key)
{
	int	x1, x2, i, n, l;
	
	/*
	// this hack is for checking the code for a key when it's pressed:
	sprintf(e->text, "key = %d", key);
	yzDrawEditline(e);
	return;
	*/
	
	caret_vis = 1;
	next_blink = centiseconds() + 40;
	if(key>255)handle_editline_cursor_key(e, key);
	else if(key != -1)	// no unknown keys in our fields pleez!
	{
		if(e->selins != e->selto)
		{
			//delete all selected characters
			x1 = min(e->selins, e->selto);
			x2 = max(e->selins, e->selto);
			n = x2 - x1;
			l = strlen(e->text);
			for(i=x1;i<=(l-n);i++)			//this should get the null terminator too
				e->text[i] = e->text[i+n];
			
			e->selto = e->selins = x1;
			
			yzDrawEditline(e);	//this requires serious display updating :P
			draw_editline_selection(e);	//don't forget the caret
			
			if(key==8 || key==127)return;	//if it was delete, we're all done
		}
		else if(key==8 || key==127)	//we have to delete a character
		{
			if(key==8)	//for a backwards delete we need to backup one
			{
				if(e->selins==0)return;	//no room to delete
				e->selins--; e->selto--;	//back up one
			}
			if(key==127)	//make sure there's something to delete
			{
				if(e->selins>=strlen(e->text))return;
			}
			l = strlen(e->text);
			for(i=e->selins;i<=(l-1);i++)	//this should get the null terminator too
				e->text[i] = e->text[i+1];
			
			yzDrawEditline(e);	//this requires serious display updating :P
			draw_editline_selection(e);	//don't forget the caret
			return;
		}
		
		if(strlen(e->text)<500)	//we have room
		{
			//insert a character to make room for the key we pressed
			l = strlen(e->text);
			for(i=l;i>=e->selins;i--)	//this should get the null terminator too
				e->text[i+1] = e->text[i];
			e->text[e->selins] = key;
			e->selins++; e->selto++;
			
			yzDrawEditline(e);	//this requires serious display updating :P
			draw_editline_selection(e);	//don't forget the caret
			return;
		}
	}
}

void	handle_editline_cursor_key(editline_t	*e, int	key)
{
	if(!yzQueryKey(YZ_KEYCODE_SHIFT))
	{
		switch(key)
		{
			case	YZ_KEYCODE_RIGHT_ARROW:
				erase_editline_selection(e);
				if(e->selins != e->selto)e->selins = e->selto = max(e->selins, e->selto);
				else if(e->selins < strlen(e->text)) { e->selins++; e->selto++; }
				draw_editline_selection(e);
				break;
			case	YZ_KEYCODE_LEFT_ARROW:
				erase_editline_selection(e);
				if(e->selins != e->selto)e->selins = e->selto = min(e->selins, e->selto);
				else if(e->selins > 0) { e->selins--; e->selto--; }
				draw_editline_selection(e);
				break;
			case	YZ_KEYCODE_UP_ARROW:
				erase_editline_selection(e);
				e->selins = e->selto = 0;
				draw_editline_selection(e);
				break;
			case	YZ_KEYCODE_DOWN_ARROW:
				erase_editline_selection(e);
				e->selins = e->selto = strlen(e->text);
				draw_editline_selection(e);
				break;
		}
	}
	else
	{
		switch(key)
		{
			case	YZ_KEYCODE_RIGHT_ARROW:
				if(e->selto < strlen(e->text))e->selto++;
				else break;
				if(e->selto > e->selins)
				{
					//draw selection background on newly selected character
					yzPenColor(50000, 50000, 65535);
					draw_editline_selected_text(e, e->selto-1, e->selto);
				}
				else
				{
					//erase selection background on newly deselected character
					yzPenColor(65535, 65535, 65535);
					draw_editline_selected_text(e, e->selto-1, e->selto);
					//draw the caret if necessary
					if(e->selto == e->selins)draw_editline_selection(e);
				}
				break;
			case	YZ_KEYCODE_LEFT_ARROW:
				if(e->selto > 0)e->selto--;
				else break;
				if(e->selto < e->selins)
				{
					//draw selection background on newly selected character
					yzPenColor(50000, 50000, 65535);
					draw_editline_selected_text(e, e->selto, e->selto+1);
				}
				else
				{
					//erase selection background on newly deselected character
					yzPenColor(65535, 65535, 65535);
					draw_editline_selected_text(e, e->selto, e->selto+1);
					//draw the caret if necessary
					if(e->selto == e->selins)draw_editline_selection(e);
				}
				break;
			case	YZ_KEYCODE_UP_ARROW:
				erase_editline_selection(e);
				e->selto = 0;
				draw_editline_selection(e);
				break;
			case	YZ_KEYCODE_DOWN_ARROW:
				erase_editline_selection(e);
				e->selto = strlen(e->text);
				draw_editline_selection(e);
				break;
		}
	}
}

int	point_in_editline(int	x, int	y, editline_t	*e)
{
	if(x<e->x || x>(e->x+e->dx-1) || y<e->y || y>(e->y+e->dy-1))return 0;
	else return 1;
}

int	editline_point_to_offset(editline_t	*e, int	x, int	y)
{
	int	i, size = 0;
	char	*txt;
	
	txt = e->do_hide?passwdtxt:e->text;
	if(y>(e->y + e->dy - 1))return strlen(txt);
	if(y<e->y)return 0;
	x -= (5 + e->x);
	for(i=0;e->text[i];i++)
	{
		if((size + yzLetterSize(txt[i])/2) > x)return i;
		size += yzLetterSize(txt[i]);
	}
	return i;
}

void	handle_editline_click(editline_t	*e, int	x, int	y)
{
	int	q, to;
	
	if(!yzQueryKey(YZ_KEYCODE_SHIFT))
	{
		erase_editline_selection(e);
		e->selto = e->selins = editline_point_to_offset(e, x, y);
		draw_editline_selection(e);
	}
	else
	{
		if(!caret_vis)draw_editline_selection(e);
		goto mouse_moved;
	}
	
	while((q=yzQueryPointer(&x, &y))>-1 && q<3)
	{
		if(q==1)	//mouse moved
		{
			mouse_moved:
			to = editline_point_to_offset(e, x, y);
			if(to == e->selto)continue;	//nothing interesting
			
			if(to > e->selto)
			{
				if(e->selto >= e->selins)	//we added forward
				{
					addforward:
					yzPenColor(50000, 50000, 65535);
					draw_editline_selected_text(e, e->selto, to);
					e->selto = to;
				}
				else	//we subtracted forward (and maybe added forward too)
				{
					yzPenColor(65535, 65535, 65535);
					draw_editline_selected_text(e, e->selto, to);
					if(to == e->selins)
					{
						e->selto = to;
						draw_editline_selection(e);
					}
					if(to > e->selins)
					{
						e->selto = e->selins;
						goto addforward;
					}
					e->selto = to;
				}
			}
			else
			{
				if(e->selto >= e->selins)	//we subtracted backwards (and maybe added back)
				{
					yzPenColor(65535, 65535, 65535);
					draw_editline_selected_text(e, to, e->selto);
					if(to == e->selins)
					{
						e->selto = to;
						draw_editline_selection(e);
					}
					if(to < e->selins)
					{
						e->selto = e->selins;
						goto addbackward;
					}
					e->selto = to;
				}
				else	//we added backwards
				{
					addbackward:
					yzPenColor(50000, 50000, 65535);
					draw_editline_selected_text(e, to, e->selto);
					e->selto = to;
				}
			}
		}
	}
	
	caret_vis= 1;
	next_blink = centiseconds() + 25;
}

draw_editline_selection(editline_t	*e)
{
	int	xoff;
	char	*txt;
	
	txt = e->do_hide?passwdtxt:e->text;
	if(e->selins == e->selto)	//single black line
	{
		yzPenColor(0, 0, 0);
		xoff = 5 + yzStringSize(txt, e->selins);
		yzLine(e->x + xoff - 1, e->y + 3, e->x + xoff - 1, e->y + e->dy - 4);
	}
	else	//selected text
	{
		yzPenColor(50000, 50000, 65535);	//light blue selection color
		draw_editline_selected_text(e, min(e->selins, e->selto),
									   max(e->selins, e->selto));
	}
}

grey_editline_selection(editline_t	*e)
{
	int	xoff;
	char	*txt;
	
	txt = e->do_hide?passwdtxt:e->text;
	if(e->selins == e->selto)	//single light grey line
	{
		yzPenColor(55000, 55000, 55000);
		xoff = 5 + yzStringSize(txt, e->selins);
		yzLine(e->x + xoff - 1, e->y + 3, e->x + xoff - 1, e->y + e->dy - 4);
	}
	else	//selected text
	{
		yzPenColor(55000, 55000, 55000);	//light grey selection color
		draw_editline_selected_text(e, min(e->selins, e->selto),
									   max(e->selins, e->selto));
	}
}

void	erase_editline_selection(editline_t	*e)
{
	int	xoff;
	char	*txt;
	
	txt = e->do_hide?passwdtxt:e->text;
	if(e->selins == e->selto)	//single white line
	{
		yzPenColor(65535, 65535, 65535);
		xoff = 5 + yzStringSize(txt, e->selins);
		yzLine(e->x + xoff - 1, e->y + 3, e->x + xoff - 1, e->y + e->dy - 4);
	}
	else	//selected text
	{
		yzPenColor(65535, 65535, 65535);	//white selection color
		draw_editline_selected_text(e, min(e->selins, e->selto),
									   max(e->selins, e->selto));
	}
}

draw_editline_selected_text(editline_t	*e, int	start, int	end)
{
	int	xoff1, xoff2;
	char	*txt;
	
	txt = e->do_hide?passwdtxt:e->text;
	xoff1 = 5 + yzStringSize(txt, start);
	xoff2 = 5 + yzStringSize(txt, end);
	yzRect(e->x + xoff1 - 1, e->y + 3, e->x + xoff2 - 1, e->y + e->dy - 4);
	yzPenColor(0, 0, 0);	//still black text
	yzText2i1p1i(e->x + xoff1, e->y + e->dy - 7, txt + start, end - start);
}

void	yzDrawEditline(editline_t	*e)
{
	yzPenColor(65535, 65535, 65535);
	yzRect(e->x, e->y, e->x + e->dx - 1, e->y + e->dy - 1);
	yzPenColor(0, 0, 0);
	yzOutlineRect(e->x, e->y, e->x + e->dx - 1, e->y + e->dy - 1);
	if(!e->do_hide)yzText(e->x + 5, e->y + e->dy - 7, e->text);
	else yzText2i1p1i(e->x + 5, e->y + e->dy - 7, passwdtxt, strlen(e->text));
}

void		yzSetEditlineMode(editline_t	*e, int	mode)
{
	e->do_hide = mode;
	yzDrawEditline(e);
}

void		yzSetEditlineText(editline_t	*e, char	*txt)
{
	strcpy(e->text, txt);
	yzDrawEditline(e);
}
